@@ -1,5 +1,5 @@
 package Thrift::API::HiveClient2;
-$Thrift::API::HiveClient2::VERSION = '0.019';
+$Thrift::API::HiveClient2::VERSION = '0.020';
 {
   $Thrift::API::HiveClient2::DIST = 'Thrift-API-HiveClient2';
 }
@@ -17,6 +17,38 @@ use Thrift;
 use Thrift::Socket;
 use Thrift::BufferedTransport;
 
+# An unclean workaround to have the multiple "die new TException" statements
+# display something useful instead of a useless stringified hash reference.
+# Currently tied to version 0.7 to 0.9; Thrift people don't seem to read cpan
+# bug reports, will need to poke them directly.
+# see https://rt.cpan.org/Public/Bug/Display.html?id=86679
+
+package Thrift::TException;
+$Thrift::TException::VERSION = '0.020';
+{
+  $Thrift::TException::DIST = 'Thrift-API-HiveClient2';
+}
+use version;
+
+my $thrift_version = version->parse($Thrift::VERSION);
+if ( $thrift_version >= version->parse('0.7.0') && $thrift_version <= version->parse('0.9.0') )
+{
+    eval <<'OVERLOAD';    #<<<
+    use overload '""' => sub {
+        return
+            ref( $_[0] )
+            . " error: "
+            . ( $_[0]->{message} || 'empty message' )
+            . " (code "
+            . ( defined $_[0]->{code} ? $_[0]->{code} : 'undefined' ) . ")";
+    };
+OVERLOAD
+}
+
+#>>>
+
+package Thrift::API::HiveClient2;
+
 # Protocol loading is done dynamically later.
 
 use Thrift::API::HiveClient2::TCLIService;
@@ -49,7 +81,7 @@ my @tabledesc_fields = qw(
     TABLE_NAME
     TABLE_TYPE
     REMARKS
-    );
+);
 
 # Don't use XS for now, fails initializing properly with BufferedTransport. See
 # Thrift::XS documentation.
@@ -79,7 +111,7 @@ has sasl => (
 
 has timeout => (
     is      => 'rw',
-    default => sub { 3_600 },
+    default => sub {3_600},
 );
 
 # These exist to make testing with various other Thrift Implementation classes
@@ -99,7 +131,7 @@ sub _set_protocol  { $_[0]->{_protocol}  = $_[1] }
 sub _set_client    { $_[0]->{_client}    = $_[1] }
 
 sub _set_sasl {
-    my ($self, $sasl) = @_;
+    my ( $self, $sasl ) = @_;
     return if !$sasl;
 
     # This normally selects XS first (hopefully)
@@ -109,11 +141,11 @@ sub _set_sasl {
     require Thrift::SASL::Transport;
     Thrift::SASL::Transport->import;
 
-    if ($sasl == 1) {
+    if ( $sasl == 1 ) {
         return $self->{_sasl} = Authen::SASL->new( mechanism => 'GSSAPI' );
     }
-    elsif (reftype $sasl eq "HASH") {
-        return $self->{_sasl} = Authen::SASL->new( %$sasl ); #, debug => 8 );
+    elsif ( reftype $sasl eq "HASH" ) {
+        return $self->{_sasl} = Authen::SASL->new(%$sasl);    #, debug => 8 );
     }
     die "Incorrect parameter passed to _set_sasl";
 }
@@ -125,8 +157,9 @@ sub BUILD {
 
     $self->_set_socket( Thrift::Socket->new( $self->host, $self->port ) )
         unless $self->_socket;
+    $self->_socket->setRecvTimeout( $self->timeout * 1000 );
 
-    $self->_set_sasl($self->sasl) if ( $self->sasl && !$self->_sasl );
+    $self->_set_sasl( $self->sasl ) if ( $self->sasl && !$self->_sasl );
 
     if ( !$self->_transport ) {
         my $transport = Thrift::BufferedTransport->new( $self->_socket );
@@ -145,7 +178,6 @@ sub BUILD {
         unless $self->_client;
 }
 
-
 sub _init_protocol {
     my $self = shift;
     my $err;
@@ -162,18 +194,17 @@ sub _init_protocol {
     return $protocol;
 }
 
-
 sub connect {
     my ($self) = @_;
-    $self->_socket->setRecvTimeout($self->timeout * 1000);
     $self->_transport->open;
 }
 
 has _session => (
-    is      => 'rwp',
-    isa     => sub {
+    is  => 'rwp',
+    isa => sub {
         die "Session isn't a Thrift::API::HiveClient2::TOpenSessionResp"
-            if ! blessed($_[0]) || !$_[0]->isa('Thrift::API::HiveClient2::TOpenSessionResp') },
+            if !blessed( $_[0] ) || !$_[0]->isa('Thrift::API::HiveClient2::TOpenSessionResp');
+    },
     lazy    => 1,
     builder => '_build_session',
 );
@@ -203,10 +234,11 @@ sub _build_session {
 }
 
 has _session_handle => (
-    is      => 'rwp',
-    isa     => sub {
+    is  => 'rwp',
+    isa => sub {
         die "Session handle isn't a Thrift::API::HiveClient2::TSessionHandle"
-            if ! blessed($_[0]) || !$_[0]->isa('Thrift::API::HiveClient2::TSessionHandle') },
+            if !blessed( $_[0] ) || !$_[0]->isa('Thrift::API::HiveClient2::TSessionHandle');
+    },
     lazy    => 1,
     builder => '_build_session_handle',
 );
@@ -232,13 +264,14 @@ has _operation => (
 );
 
 has _operation_handle => (
-    is => 'rwp',
+    is  => 'rwp',
     isa => sub {
-        die
-            "Operation handle isn't a Thrift::API::HiveClient2::TOperationHandle"
-            if defined $_[0] && ( ! blessed($_[0]) || !$_[0]->isa('Thrift::API::HiveClient2::TOperationHandle') );
+        die "Operation handle isn't a Thrift::API::HiveClient2::TOperationHandle"
+            if defined $_[0]
+            && ( !blessed( $_[0] )
+            || !$_[0]->isa('Thrift::API::HiveClient2::TOperationHandle') );
     },
-    lazy    => 1,
+    lazy => 1,
 );
 
 sub _cleanup_previous_operation {
@@ -248,16 +281,15 @@ sub _cleanup_previous_operation {
     # operation handle explicitely
     if ( $self->_operation_handle ) {
         $self->_client->CloseOperation(
-             Thrift::API::HiveClient2::TCloseOperationReq->new(
-                 { operationHandle => $self->_operation_handle, }
-             )
+            Thrift::API::HiveClient2::TCloseOperationReq->new(
+                { operationHandle => $self->_operation_handle, }
+            )
         );
         $self->_set__operation(undef);
         $self->_set__operation_handle(undef);
     }
 }
 
-
 sub execute {
     my $self = shift;
     my ($query) = @_;    # make this a bit more flexible
@@ -269,42 +301,47 @@ sub execute {
             { sessionHandle => $self->_session_handle, statement => $query, confOverlay => {} }
         )
     );
-    if ($rh->{status}{errorCode}) {
+    if ( $rh->{status}{errorCode} ) {
         die __PACKAGE__ . "::execute: $rh->{status}{errorMessage}; HQL was: \"$query\"";
     }
     $self->_set__operation($rh);
-    $self->_set__operation_handle($rh->{operationHandle});
+    $self->_set__operation_handle( $rh->{operationHandle} );
     return $rh;
 }
 
 {
     # cache the column names we need to extract from the bloated data structure
     # (keyed on query)
-    my ($column_keys, $column_names);
+    my ( $column_keys, $column_names );
 
     sub fetch_hashref {
         my $self = shift;
-        my ( $rv, $rows_at_a_time ) = @_;
-        return $self->fetch($rv, $rows_at_a_time, 1);
+        my ( $rh, $rows_at_a_time ) = @_;
+        return $self->fetch( $rh, $rows_at_a_time, 1 );
     }
 
     sub fetch {
         my $self = shift;
-        my ( $rv, $rows_at_a_time, $use_hashref ) = @_;
+        my ( $rh, $rows_at_a_time, $use_hashref ) = @_;
 
-        # if $rv looks like a number, use it instead of $rows_at_a_time
-        $rows_at_a_time = $rv if !$rows_at_a_time && $rv =~ /^[1-9][0-9]*$/;
+        # if $rh looks like a number, use it instead of $rows_at_a_time
+        # it means we're using the new form for this call, which takes only the
+        # number of wanted rows, or even nothing (and relies on the defaults,
+        # and a cached copy of the query $rh)
+        $rows_at_a_time = $rh if ( $rh && !$rows_at_a_time && $rh =~ /^[1-9][0-9]*$/ );
+        $rows_at_a_time ||= 10_000;
 
         my $result = [];
         my $has_more_rows;
 
-        # NOTE we don't use $rv now, maybe we should leave that possibility open
-        # for parallel queries, but that woudl need a lot more testing. Patches
-        # welcome.
-        my $rh = $self->_client->FetchResults(
+        # NOTE we don't use the provided $rh any more, maybe we should leave
+        # that possibility open for parallel queries, but that would need a lot
+        # more testing. Patches welcome.
+        my $cached_rh = $self->_operation_handle;
+        $rh = $self->_client->FetchResults(
             Thrift::API::HiveClient2::TFetchResultsReq->new(
-                {   operationHandle => $self->_operation_handle,
-                    maxRows         => $rows_at_a_time || 10_000
+                {   operationHandle => $cached_rh,
+                    maxRows         => $rows_at_a_time,
                 }
             )
         );
@@ -317,17 +354,17 @@ sub execute {
             for my $row ( @{ $rh->{results}{rows} || [] } ) {
 
                 # Find which fields to extract from each row, only on the first iteration
-                if ( !@{ $column_keys->{ $rv } || [] } ) {
+                if ( !@{ $column_keys->{$cached_rh} || [] } ) {
 
                     # metadata for the query
                     if ($use_hashref) {
                         my $rh_meta = $self->_client->GetResultSetMetadata(
-                            Thrift::API::HiveClient2::TGetResultSetMetadataReq->new({
-                                operationHandle => $self->_operation_handle
-                            })
+                            Thrift::API::HiveClient2::TGetResultSetMetadataReq->new(
+                                { operationHandle => $cached_rh }
+                            )
                         );
-                        $column_names =
-                            [ map { $_->{columnName} } @{ $rh_meta->{schema}{columns} || [] } ];
+                        $column_names = [ map { $_->{columnName} }
+                                @{ $rh_meta->{schema}{columns} || [] } ];
                     }
 
                     # TODO redo all this using the TGetResultSetMetadataResp object we retrieved
@@ -344,18 +381,18 @@ sub execute {
                         # NOTE this data structure smells of Java and friends from
                         # miles away. Dynamically typed languages don't really need
                         # the bloat.
-                        push @{ $column_keys->{$rv} }, grep { ref $first_col->{$_} } keys %$first_col;
+                        push @{ $column_keys->{$cached_rh} },
+                            grep { ref $first_col->{$_} } keys %$first_col;
                     }
                 }
 
                 # TODO find something faster? (see comment above)
 
-                my $idx = 0;
+                my $idx    = 0;
                 my $retval = [
-                    map { $_->value }
-                        grep { defined $_ }
-                            map  { $row->{colVals}[ $idx++ ]{$_} }
-                                @{ $column_keys->{$rv} }
+                    map  { $_->value }
+                    grep { defined $_ }
+                    map  { $row->{colVals}[ $idx++ ]{$_} } @{ $column_keys->{$cached_rh} }
                 ];
                 if ($use_hashref) {
                     push @$result, { zip @$column_names, @$retval };
@@ -422,8 +459,7 @@ sub get_tables {
 
     my $rh = $self->_client->GetTables(
         Thrift::API::HiveClient2::TGetTablesReq->new(
-            {
-                sessionHandle => $self->_session_handle,
+            {   sessionHandle => $self->_session_handle,
                 catalogName   => undef,
                 schemaName    => $schema,
                 tableName     => $table_pattern,
@@ -477,7 +513,6 @@ sub AUTOLOAD {
     croak "No such method exists: $AUTOLOAD";
 }
 
-
 1;
 
 __END__
@@ -492,7 +527,7 @@ Thrift::API::HiveClient2 - Perl to HiveServer2 Thrift API wrapper
 
 =head1 VERSION
 
-version 0.019
+version 0.020
 
 =head1 METHODS
 
@@ -623,6 +658,10 @@ L<https://groups.google.com/a/cloudera.org/d/msg/cdh-user/AXeEuaFP0Ro/Txmn1OHleA
 So we had to change the init script for hive-server2 to make it behave, adding
 '-Dfile.encoding=UTF-8' to HADOOP_OPTS
 
+=head1 REPOSITORY
+
+L<https://github.com/dmorel/Thrift-API-HiveClient2>
+
 =head1 CONTRIBUTORS
 
 Burak Gürsoy (BURAK)
